// Copyright (C) 1999-2000 Open Source Telecom Corporation.
//  
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software 
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
// 
// As a special exception to the GNU General Public License, permission is 
// granted for additional uses of the text contained in its release 
// of ccscript.
// 
// The exception is that, if you link the ccscript library with other
// files to produce an executable, this does not by itself cause the
// resulting executable to be covered by the GNU General Public License.
// Your use of that executable is in no way restricted on account of
// linking the ccscript library code into it.
// 
// This exception does not however invalidate any other reasons why
// the executable file might be covered by the GNU General Public License.
// 
// This exception applies only to the code released under the 
// name ccscript.  If you copy code from other releases into a copy of
// ccscript, as the General Public License permits, the exception does
// not apply to the code that you add in this way.  To avoid misleading
// anyone as to the status of such modified files, you must delete
// this exception notice from them.
// 
// If you write modifications of your own for ccscript, it is your choice
// whether to permit this exception to apply to your modifications.
// If you do not wish that, delete this exception notice.  

#include "private.h"

RTPQueue::RTPQueue(int pri) :
Thread(NULL, pri) 
{
	sendfirst = sendlast = recvfirst = recvlast = NULL;
	sendseq = recvseq = 0;
	sendsources[0] = htonl(getpid());
	sendcc = 0;
	segment = 8000;
	active = false;
	timeout = 8;
	expired = 40;
} 

void RTPQueue::endQueue(void)
{
	RTPPacket *packet, *next;

	active = false;
	Terminate();
	sendlock.EnterMutex();	
	recvlock.EnterMutex();

	packet = sendfirst;
	while(packet)
	{
		next = packet->next;
		delete packet;
		packet = next;
	}

	packet = recvfirst;
	while(packet)
	{
		next = packet->next;
		delete packet;
		packet = next;
	}

	recvlock.LeaveMutex();
	sendlock.LeaveMutex();
}

timeout_t RTPQueue::getTimeout(void)
{
	struct timeval send, now;
	long diff;
	unsigned long rate;
	unsigned long stamp;
	unsigned long rem;
	RTPPacket *packet;

	for(;;)
	{
		if(!sendfirst)
			return timeout;

		stamp = ntohl(sendfirst->getHeader()->timestamp);
		switch(sendfirst->getHeader()->payload)
		{
		case RTP_PAYLOAD_DVI4_16000:
			rate = 16000;
			break;
		case RTP_PAYLOAD_L16_DUAL:
		case RTP_PAYLOAD_L16_MONO:
			rate = 44100;
			break;
		case RTP_PAYLOAD_DVI4_11025:
			rate = 11025;
			break;
		case RTP_PAYLOAD_DVI4_22050:
			rate = 22050;
			break;
		case RTP_PAYLOAD_NV:
		case RTP_PAYLOAD_CELB:
		case RTP_PAYLOAD_H261:
		case RTP_PAYLOAD_MPV:
		case RTP_PAYLOAD_MP2T:
		case RTP_PAYLOAD_H263:
			rate = 90000;
			break;
		default:
			rate = 8000;
		}

		send.tv_sec = stamp / rate;
		rem = stamp % rate;
		send.tv_usec = 1000000l * rem / rate;
		gettimeofday(&now, NULL);
	
		send.tv_sec += starttimer.tv_sec;
		send.tv_usec += starttimer.tv_usec;
		if(send.tv_usec >= 1000000l)
		{
			++send.tv_sec;	
			send.tv_usec %= 1000000l;
		}
	
		if(send.tv_sec - now.tv_sec > 3600)
			return 3600000l;

		diff = (send.tv_sec - now.tv_sec) * 1000;
		diff += (send.tv_usec - now.tv_usec) / 1000l;
		if(diff >= 0)
			return (timeout_t)diff;

		if((diff < 0) && ((timeout_t)(-diff) < expired))
			return 0;

		setCancel(THREAD_CANCEL_DEFERRED);
		sendlock.EnterMutex();
		packet = sendfirst->next;
		delete sendfirst;
		if(packet)
			packet->prev = NULL;
		sendfirst = packet;
		if(!sendfirst)
			sendlast = NULL;
		sendlock.LeaveMutex();		
		setCancel(THREAD_CANCEL_IMMEDIATE);
	}
	return diff;
}

void RTPQueue::putPacket(unsigned long stamp, payload_t payload, unsigned char *data, unsigned len)
{
	unsigned char buffer[sizeof(RTPHeader) + 64];
	RTPHeader *header = (RTPHeader *)&buffer;
	RTPPacket *packet;

	header->payload = payload;
	header->sequence = htons(sendseq);
	header->timestamp = htonl(stamp);
	header->cc = sendcc;
	header->version = 2;
	header->padding = 0;
	header->extension = 0;
	header->marker = 0;
	memcpy(&header->sources, &sendsources, 4 + 16 * sendcc);

	if(data && len)
		++sendseq;
	
	packet = new RTPPacket(header, data, len);
	sendlock.EnterMutex();
	packet->prev = sendlast;
	if(sendlast)
		sendlast->next = packet;
	sendlast = packet;
	if(!sendfirst)
		sendfirst = packet;
	sendlock.LeaveMutex();
}

int RTPQueue::recvPacket(void)
{
	int rtn;
	unsigned short seq;
	RTPHeader *header = (RTPHeader *)&buffer;
	RTPPacket *list, *next, *packet;
	int hdrsize;

	rtn = readPacket(buffer, sizeof(buffer));
	if(rtn < 0)
		return rtn;

	hdrsize = sizeof(RTPHeader) + 4 * header->cc;
	if(header->padding && rtn > hdrsize)
	{
		rtn -= buffer[rtn - 1];
		header->padding = 0;
	}

	rtn = rtn - hdrsize;
	if(rtn < 1)
		return 0;

	seq = ntohs(header->sequence);
	packet = new RTPPacket(header, buffer + hdrsize, rtn);
	setCancel(THREAD_CANCEL_DEFERRED);
	recvlock.EnterMutex();
	list = recvfirst;
	while(list)
	{
		if(seq > ntohs(list->getHeader()->sequence))
		{
			list = list->next;
			continue;
		}
		break;
	}
	if(!list)
	{
		if(recvlast)
			recvlast->next = packet;
		packet->prev = recvlast;
		recvlast = packet;
		if(!recvfirst)
			recvfirst = packet;
	}
	else if(list == recvfirst)
	{
		recvfirst->prev = packet;
		packet->next = recvfirst;
		recvfirst = packet;
	}
	else
	{
		list->prev->next = packet;
		packet->prev = list->prev;
		packet->next = list;
		list->prev = packet;
	}

	recvlock.LeaveMutex();
	return rtn;
}
	

int RTPQueue::sendPacket(void)
{
	int rtn;

	RTPPacket *packet = sendfirst;
	if(!packet)
		return 0;
	
	rtn = packet->size;
	if(rtn)
		rtn = writePacket(packet->buffer, rtn + packet->getHeaderSize());		

	if(rtn > -1)
	{
		setCancel(THREAD_CANCEL_DEFERRED);
		sendlock.EnterMutex();
		sendfirst = packet->next;
		if(sendfirst)
			sendfirst->prev = NULL;
		if(!sendfirst)
			sendlast = NULL;
		delete packet;
	}
	return rtn;
}

RTPPacket *RTPQueue::getWaiting(unsigned long stamp)
{
	unsigned long pstamp;
	RTPHeader *header;
	RTPPacket *packet;
	recvlock.EnterMutex();
	packet = recvfirst;
	while(packet)
	{
		header = packet->getHeader();
		pstamp = ntohl(header->timestamp);
		if(pstamp < stamp)
		{
			if(packet->next)
				packet->next->prev = NULL;
			else
				recvlast = packet->prev;
			recvfirst = packet->next;
			delete packet;
			packet = recvfirst;
			continue;
		}
		if(pstamp > stamp)
			packet = NULL;
		break;
	}
	recvlock.LeaveMutex();
	return packet;
}

unsigned RTPQueue::setPartial(unsigned long stamp, unsigned char *data, unsigned offset, unsigned max)
{
	RTPPacket *packet;
	unsigned long pstamp;

	sendlock.EnterMutex();
	packet = sendfirst;
	while(packet)
	{
		pstamp = ntohl(packet->getHeader()->timestamp);
		if(pstamp > stamp)
			packet = NULL;
		if(pstamp >= stamp)
			break;

		packet = packet->next;
	}
	if(!packet)
	{
		sendlock.LeaveMutex();
		return 0;
	}

	if(offset >= packet->getPayloadSize())
		return 0;

	if(max > packet->getPayloadSize() - offset)
		max = packet->getPayloadSize() - offset;

	memcpy(packet->getPayload() + offset, data, max);
	sendlock.LeaveMutex();
	return max;
}

unsigned RTPQueue::getPartial(unsigned long stamp, unsigned char *data, unsigned offset, unsigned max)
{
	RTPPacket *packet = getWaiting(stamp);

	if(!packet)
		return 0;

	if(offset >= packet->getPayloadSize())
		return 0;

	if(max > packet->size - offset)
		max = packet->size - offset;

	memcpy(data, packet->getPayload() + offset, max);
	return max;
}
		
unsigned RTPQueue::getPacket(unsigned long stamp, unsigned char *data, unsigned max)
{
	RTPPacket *packet;
	RTPHeader *header;
	unsigned count = 0;
	int len;
	unsigned short seq;

	complete = true;
	while(NULL != (packet = getWaiting(stamp)))
	{
		len = packet->size;
		if(max < len)
			len = max;

		memcpy(data, packet->getPayload(), len);
		recvlock.EnterMutex();
		if(!packet->prev)
			recvfirst = packet->next;
		else
			packet->prev->next = packet->next;
		if(!packet->next)
			recvlast = packet->prev;
		else
			packet->next->prev = packet->prev;
		header = packet->getHeader();
		marked = header->marker;
		seq = ntohs(header->sequence);
		if(recvseq != seq)
			complete = false;
		recvseq = ++seq;
		delete packet;
		recvlock.LeaveMutex();
		count += len;
		data += len;
		max -= len;
		if(max < 1)
			break;
	}
	return count;
}

payload_t RTPQueue::getPayload(unsigned long stamp)
{
	RTPPacket *packet = getWaiting(stamp);

	if(!packet)
		return RTP_PAYLOAD_INVALID;

	return (payload_t)(packet->getHeader()->payload);
}

void	RTPQueue::Run(void)
{
	timeout_t wait = 0;
	time_t control, now, check, rtcp;

	time(&control);
	time(&check);
	while(active)
	{
		setCancel(THREAD_CANCEL_IMMEDIATE);
		if(!wait)
			wait = getTimeout();

		rtcp = getControlTimer();
		if(rtcp)
		{
			time(&now);
			if(wait > 1000)
				wait = 1000;

			if(check != now)	
			{
				getControl();
				check = now;
			}
			if(control + rtcp <= now)
			{
				putControl();
				control = now;
			}
		}
		if(!wait)
		{
			if(sendPacket() < 0)
				wait = timeout;			
			continue;
		}
		setCancel(THREAD_CANCEL_IMMEDIATE);
		if(isPendingArrival(wait))
			if(recvPacket() < 0)
				break;
		wait = 0;
	}
	active = false;
	Bye();
	Sleep(~0);
}


